import attrs
import re
from pydra.compose import base


@attrs.define(kw_only=True)
class arg(base.Arg):
    """Argument of a Python task

    Parameters
    ----------
    help: str
        A short description of the input field.
    default : Any, optional
        the default value for the argument
    allowed_values: list, optional
        List of allowed values for the field.
    requires: list, optional
        Names of the inputs that are required together with the field.
    copy_mode: File.CopyMode, optional
        The mode of copying the file, by default it is File.CopyMode.any
    copy_collation: File.CopyCollation, optional
        The collation of the file, by default it is File.CopyCollation.any
    copy_ext_decomp: File.ExtensionDecomposition, optional
        The extension decomposition of the file, by default it is
        File.ExtensionDecomposition.single
    readonly: bool, optional
        If True the input field can’t be provided by the user but it aggregates other
        input fields (for example the fields with argstr: -o {fldA} {fldB}), by default
        it is False
    type: type, optional
        The type of the field, by default it is Any
    name: str, optional
        The name of the field, used when specifying a list of fields instead of a mapping
        from name to field, by default it is None
    path: str | None
        The path to where the input is stored within the dataset: the modality of the
        file followed by the BIDS suffix in the form 'modality/suffix'.
    """

    # the BIDS suffix that identifies the input
    path: str | None = attrs.field()

    @path.validator
    def _path_validator(self, attribute, value):
        """Validate the path of the input field"""
        if value is None:
            return
        if not isinstance(value, str):
            raise TypeError(f"Path must be a string, got {type(value)}")
        if not re.match(r"\w+/\w+", value):
            raise ValueError(f"Path must be of the form 'modality/suffix', got {value}")


@attrs.define(kw_only=True)
class out(base.Out):
    """Output of a Python task

    Parameters
    ----------
    name: str, optional
        The name of the field, used when specifying a list of fields instead of a mapping
        from name to field, by default it is None
    type: type, optional
        The type of the field, by default it is Any
    help: str, optional
        A short description of the input field.
    requires: list, optional
        Names of the inputs that are required together with the field.
    converter: callable, optional
        The converter for the field passed through to the attrs.field, by default it is None
    validator: callable | iterable[callable], optional
        The validator(s) for the field passed through to the attrs.field, by default it is None
    position : int
        The position of the output in the output list, allows for tuple unpacking of
        outputs
    path: str, optional
        The path to where the output is stored within the derivatives directory, the
        namespace of the derivative followed by suffix of the output file in the
        form 'modality/suffix'.
    """

    # the path the file will be stored within the deriavtives directory
    path: str = attrs.field(default="")

    @path.validator
    def _path_validator(self, attribute, value):
        """Validate the path of the input field"""
        if not isinstance(value, str):
            raise TypeError(f"Path must be a string, got {type(value)}")



import os
import pytest
import requests
from warnings import warn
import typing as ty
from tempfile import mkdtemp
from pathlib import Path
import docker
from fileformats.medimage import NiftiGzX

PKG_DIR = Path(__file__).parent


@pytest.fixture(scope="session")
def build_cache_dir() -> Path:
    # build_cache_dir = Path.home() / '.frametree-test-build-cache'
    # if build_cache_dir.exists():
    #     shutil.rmtree(build_cache_dir)
    # build_cache_dir.mkdir()
    return Path(mkdtemp())
    # return build_cache_dir


@pytest.fixture
def work_dir() -> Path:  # type: ignore[misc]
    # work_dir = Path.home() / '.frametree-tests'
    # work_dir.mkdir(exist_ok=True)
    # return work_dir
    work_dir = mkdtemp()
    yield Path(work_dir)
    # shutil.rmtree(work_dir)


@pytest.fixture(scope="session")
def pkg_dir() -> Path:
    return PKG_DIR


@pytest.fixture(scope="session")
def nifti_sample_dir(pkg_dir):
    return pkg_dir / "test-data" / "nifti"


# FIXME: should be converted to python script to be Windows compatible
@pytest.fixture(scope="session")
def mock_bids_app_script():
    file_tests = ""
    for inpt_path, datatype in [
        ("anat/T1w", NiftiGzX),
        ("anat/T2w", NiftiGzX),
        ("dwi/dwi", NiftiGzX),
    ]:
        subdir, suffix = inpt_path.split("/")
        fpath = (
            f"$BIDS_DATASET/sub-${{SUBJ_ID}}/{subdir}/"
            f"sub-${{SUBJ_ID}}_{suffix}{datatype.ext}"
        )
        file_tests += f"""
        if [ ! -f {fpath} ]; then
            echo "Did not find {suffix} file at {fpath}"
            exit 1;
        fi
        """

    return f"""#!/bin/sh
BIDS_DATASET=$1
OUTPUTS_DIR=$2
SUBJ_ID=$5
{file_tests}
# Write mock output files to 'derivatives' Directory
mkdir -p $OUTPUTS_DIR
echo 'file1' > $OUTPUTS_DIR/sub-${{SUBJ_ID}}_file1.txt
echo 'file2' > $OUTPUTS_DIR/sub-${{SUBJ_ID}}_file2.txt
"""


# For debugging in IDE's don't catch raised exceptions and let the IDE
# break at it
if os.getenv("_PYTEST_RAISE", "0") != "0":

    @pytest.hookimpl(tryfirst=True)
    def pytest_exception_interact(call: pytest.CallInfo[ty.Any]) -> None:
        if call.excinfo is not None:
            raise call.excinfo.value

    @pytest.hookimpl(tryfirst=True)
    def pytest_internalerror(excinfo: pytest.ExceptionInfo[BaseException]) -> None:
        raise excinfo.value

    CATCH_CLI_EXCEPTIONS = False
else:
    CATCH_CLI_EXCEPTIONS = True


@pytest.fixture(scope="session")
def bids_validator_app_image(
    bids_validator_app_script, bids_validator_docker, build_cache_dir
):
    return build_app_image(
        BIDS_VALIDATOR_APP_IMAGE,
        bids_validator_app_script,
        build_cache_dir,
        base_image=bids_validator_docker,
    )


def build_app_image(tag_name, script, build_cache_dir, base_image):
    dc = docker.from_env()

    # Create executable that runs validator then produces some mock output
    # files
    build_dir = build_cache_dir / tag_name.replace(":", "__i__")
    build_dir.mkdir()
    launch_sh = build_dir / "launch.sh"
    with open(launch_sh, "w") as f:
        f.write(script)

    # Build mock BIDS app image
    with open(build_dir / "Dockerfile", "w") as f:
        f.write(
            f"""FROM {base_image}
ADD ./launch.sh /launch.sh
RUN chmod +x /launch.sh
ENTRYPOINT ["/launch.sh"]"""
        )

    dc.images.build(path=str(build_dir), tag=tag_name)

    return tag_name


@pytest.fixture(scope="session")
def bids_command_spec(mock_bids_app_executable):
    inputs = {
        "T1w": {
            "configuration": {
                "path": "anat/T1w",
            },
            "datatype": "medimage:NiftiGzX",
            "help": "T1-weighted image",
        },
        "T2w": {
            "configuration": {
                "path": "anat/T2w",
            },
            "datatype": "medimage:NiftiGzX",
            "help": "T2-weighted image",
        },
        "DWI": {
            "configuration": {
                "path": "dwi/dwi",
            },
            "datatype": "medimage:NiftiGzXBvec",
            "help": "DWI-weighted image",
        },
    }

    outputs = {
        "file1": {
            "configuration": {
                "path": "file1",
            },
            "datatype": "common:Text",
            "help": "an output file",
        },
        "file2": {
            "configuration": {
                "path": "file2",
            },
            "datatype": "common:Text",
            "help": "another output file",
        },
    }

    return {
        "task": "frametree.bids.tasks:bids_app",
        "inputs": inputs,
        "outputs": outputs,
        "row_frequency": "session",
        "configuration": {
            "inputs": inputs,
            "outputs": outputs,
            "executable": str(mock_bids_app_executable),
        },
    }


@pytest.fixture(scope="session")
def bids_validator_docker():
    dc = docker.from_env()
    try:
        dc.images.pull(BIDS_VALIDATOR_DOCKER)
    except requests.exceptions.HTTPError:
        warn("No internet connection, so couldn't download latest BIDS validator")
    return BIDS_VALIDATOR_DOCKER


@pytest.fixture(scope="session")
def bids_validator_app_script():
    return f"""#!/bin/sh
# Echo inputs to get rid of any quotes
BIDS_DATASET=$(echo $1)
OUTPUTS_DIR=$(echo $2)
SUBJ_ID=$5
# Run BIDS validator to check whether BIDS dataset is created properly
output=$(/usr/local/bin/bids-validator "$BIDS_DATASET")
if [[ "$output" != *"{SUCCESS_STR}"* ]]; then
    echo "BIDS validation was not successful, exiting:\n "
    echo $output
    exit 1;
fi
# Write mock output files to 'derivatives' Directory
mkdir -p $OUTPUTS_DIR
echo 'file1' > $OUTPUTS_DIR/sub-${{SUBJ_ID}}_file1.txt
echo 'file2' > $OUTPUTS_DIR/sub-${{SUBJ_ID}}_file2.txt
"""


BIDS_VALIDATOR_DOCKER = "bids/validator:latest"
SUCCESS_STR = "This dataset appears to be BIDS compatible"
BIDS_VALIDATOR_APP_IMAGE = "pydra-compose-bidsapp-validator"



import os
import stat
from pathlib import Path
import shutil
import pytest
from fileformats.medimage import NiftiGzX, NiftiGzXBvec
from pydra.compose import bidsapp
from pydra.utils import asdict
from fileformats.text import Plain as Text
from fileformats.generic import Directory


MOCK_BIDS_APP_NAME = "mockapp"
MOCK_README = "A dummy readme\n" * 100
MOCK_AUTHORS = ["Dumm Y. Author", "Another D. Author"]

BIDS_INPUTS = [
    bidsapp.arg(name="T1w", path="anat/T1w", type=NiftiGzX),
    bidsapp.arg(name="T2w", path="anat/T2w", type=NiftiGzX),
    bidsapp.arg(name="dwi", path="dwi/dwi", type=NiftiGzXBvec),
]
BIDS_OUTPUTS = [
    bidsapp.out(name="whole_dir", type=Directory),  # whole derivative directory
    bidsapp.out(name="a_file", path="file1", type=Text),
    bidsapp.out(name="another_file", path="file2", type=Text),
]


@pytest.mark.xfail(reason="this test will fail until nipype/pydra#822 is merged")
def test_bids_app_docker(
    bids_validator_app_image: str, nifti_sample_dir: Path, work_dir: Path
):

    bids_dir = work_dir / "bids"

    shutil.rmtree(bids_dir, ignore_errors=True)

    TestBids = bidsapp.define(
        bids_validator_app_image,
        inputs=BIDS_INPUTS,
        outputs=BIDS_OUTPUTS,
    )

    task = TestBids(
        work_dir=work_dir / "test-work",
        output_dir=work_dir / "test-outputs",
        **{
            inpt.name: nifti_sample_dir.joinpath(*inpt.path.split("/")).with_suffix(
                inpt.type.ext
            )
            for inpt in BIDS_INPUTS
        },
    )

    result = task(worker="debug")

    for output in BIDS_OUTPUTS:
        assert Path(getattr(result.output, output.name)).exists()


def test_bids_app_naked(
    mock_bids_app_script: str, nifti_sample_dir: Path, work_dir: Path
):

    # Create executable that runs validator then produces some mock output
    # files
    launch_sh = work_dir / "launch.sh"
    # We don't need to run the full validation in this case as it is already tested by test_run_bids_app_docker
    # so we use the simpler test script.
    with open(launch_sh, "w") as f:
        f.write(mock_bids_app_script)

    os.chmod(launch_sh, stat.S_IRWXU)

    TestBids = bidsapp.define(
        launch_sh,
        inputs=BIDS_INPUTS,
        outputs=BIDS_OUTPUTS,
    )

    task = TestBids(
        **{
            i.name: nifti_sample_dir.joinpath(*i.path.split("/")).with_suffix(
                i.type.ext
            )
            for i in BIDS_INPUTS
        },
    )
    outputs = task(worker="debug")

    for output in asdict(outputs).values():
        assert Path(output).exists()



import attrs
import typing as ty
from pathlib import Path
import logging
from frametree.core import __version__
from frametree.core.frameset import FrameSet
from frametree.common import Clinical
from frametree.bids.store import Bids
from pydra.utils import asdict, get_fields
from pydra.compose import base
from pydra.environments.docker import Docker
from pydra.environments.native import Native
from . import fields
from .app import BidsApp

logger = logging.getLogger("pydra.compose.bidsapp")

if ty.TYPE_CHECKING:
    from pydra.engine.job import Job


@attrs.define(kw_only=True, auto_attribs=False, eq=False, repr=False)
class BidsAppOutputs(base.Outputs):

    @classmethod
    def _from_job(cls, job: "Job[BidsAppTask]") -> ty.Self:
        """Collect the outputs of a job from a combination of the provided inputs,
        the objects in the output directory, and the stdout and stderr of the process.

        Parameters
        ----------
        job : Job[Task]
            The job whose outputs are being collected.
        outputs_dict : dict[str, ty.Any]
            The outputs of the job, as a dictionary

        Returns
        -------
        outputs : Outputs
            The outputs of the job in dataclass
        """
        outputs = super()._from_job(job)
        frameset: FrameSet = job.return_values["frameset"]
        output_fields: ty.List[fields.out] = get_fields(cls)
        for output_field in output_fields:
            if output_field.path:
                path = output_field.path
            else:
                path = ""  # whole directory
            path += "@" + DEFAULT_DERIVATIVES_NAME
            frameset.add_sink(
                output_field.name,
                output_field.type,
                path=path,
            )
        row = frameset.row(Clinical.session, DEFAULT_BIDS_ID)
        with frameset.store.connection:
            for output_field in output_fields:
                setattr(outputs, output_field.name, row[output_field.name])
        return outputs


BidsAppOutputsType = ty.TypeVar("BidsAppOutputsType", bound=BidsAppOutputs)


@attrs.define(kw_only=True, auto_attribs=False, eq=False, repr=False)
class BidsAppTask(base.Task[BidsAppOutputsType]):

    _executor_name = "app"

    BASE_ATTRS = (
        "analysis_level",
        "json_edits",
        "flags",
        "app",
        "work_dir",
    )

    analysis_level: str = fields.arg(
        name="analysis_level",
        type=str,
        default="participant",
        help="Level of analysis to run the app at",
        path=None,
    )
    json_edits: list[tuple[str, str]] | None = fields.arg(
        name="json_edits",
        type=list[tuple[str, str]] | None,
        default=None,
        path=None,
    )
    flags: str | None = fields.arg(
        name="flags",
        type=str | None,
        default=None,
        help=(
            "Additional flags to pass to the app. These are passed as a single string "
            "and should be formatted as they would be on the command line "
            "(e.g. '--flag1 --flag2 value')"
        ),
        path=None,
    )
    work_dir: Path | None = fields.arg(
        type=Path | None,
        default=None,
        help="The directory where the temporary BIDS dataset will be created and Pydra cache stored.",
        path=None,
    )

    def _run(self, job: "Job[BidsAppTask]", rerun: bool = True) -> None:
        # Create a BIDS dataset and save input data into it
        job.return_values["frameset"] = frameset = self._create_dataset()
        output_dir = (
            Path(frameset.id)
            / "derivatives"
            / DEFAULT_DERIVATIVES_NAME
            / f"sub-{DEFAULT_BIDS_ID}"
        )
        output_dir.mkdir(parents=True, exist_ok=True)
        if self.work_dir:
            work_dir = self.work_dir
        else:
            work_dir = Path.cwd() / "work-dir"
        work_dir.mkdir(parents=True, exist_ok=True)
        cache_root = work_dir / "internal-cache"
        app_work_dir = work_dir / "app-work-dir"
        cache_root.mkdir()
        app_work_dir.mkdir()

        if self.app.startswith("/"):
            executable = self.app
            image_tag = None
        else:
            if self.app.startswith("docker://"):
                image_tag = self.app.split("docker://")[-1]
            else:
                logger.info(
                    "Assuming that the wrapped string, '%s' is a docker image tag. ",
                    self.app,
                )
                image_tag = self.app
            if "::" in image_tag:
                image_tag, executable = image_tag.split("::")
            else:
                executable = None  # entrypoint of the container

        app = BidsApp(
            executable=executable,
            dataset_path=frameset.id,
            output_path=output_dir,
            analysis_level=self.analysis_level,
            participant_label=DEFAULT_BIDS_ID,
            flags=self.flags,
            work_dir=app_work_dir,
        )
        environment = Docker(image_tag) if image_tag else Native()
        app(cache_root=cache_root, environment=environment)

    def _create_dataset(self) -> FrameSet:
        # Prepare the inputs to the function
        inputs = asdict(self)
        for atr in self.BASE_ATTRS:
            del inputs[atr]

        dataset_dir = Path.cwd() / "bids-dataset"
        frameset = Bids().create_dataset(
            id=dataset_dir,
            name=DEFAULT_FRAMESET_NAME,
            leaves=[(DEFAULT_BIDS_ID,)],
            metadata={"authors": [f"Auto-generated by FrameTree {__version__}"]},
        )

        # Update the Bids store with the JSON edits requested by the user
        # je_args = shlex.split(self.json_edits) if json_edits else []
        bids_store: Bids = frameset.store
        bids_store.json_edits = self.json_edits

        # JsonEdit.attr_converter(
        #     self.json_edits + list(zip(je_args[::2], je_args[1::2]))
        # )
        input_fields: list[fields.arg] = [
            f for f in get_fields(self) if f.name not in self.BASE_ATTRS
        ]
        for inpt in input_fields:
            frameset.add_sink(inpt.name, inpt.type, path=inpt.path)
        row = frameset.row(Clinical.session, DEFAULT_BIDS_ID)
        with frameset.store.connection:
            for inpt in input_fields:
                inpt_value = inputs[inpt.name]
                if not inpt_value:
                    logger.warning("No input provided for '%s' input", inpt.name)
                    continue
                row[inpt.name] = inpt_value
        return frameset


# For running
CONTAINER_DERIV_PATH = "/frametree_bids_outputs"
CONTAINER_DATASET_PATH = "/frametree_bids_dataset"

DEFAULT_BIDS_ID = "DEFAULT"
DEFAULT_FRAMESET_NAME = "DEFAULT"
DEFAULT_DERIVATIVES_NAME = "DEFAULT"
OUTPUT_DIR_NAME = "output-dir"
